/******************************************************************************
 *
 * Copyright (c) 2019 by Ralf Burger (ralf@Burger-AG.de)
 *
 * Raw-Ethernet component driver (raspberry and  arduino)
 *
 *       LinuxCNC - Arduino-/Raspberry- IO
 *       pcIOethRaw.c  - (c) by Ralf Burger 2019 -
 *       V.: 0.93 Last changes:  29.03.2019
 *
 * Description:
 * A direct ethernet-connection to the correspondig linuxCNC-eth-device
 * is nessesarry.
 * This programm sends raw-eth-packets to a connected arduino or raspberry
 *
 * Author: Ralf Burger
 * License: GPL Version 2
 *
 * Copyright (c) 2019 All rights reserved.
 *
 *
 ******************************************************************************
 *
 * test-functions:
 * keys 1-8 = toggles bit 0-7 of output-byte
 * "+" shifts betweeen output-byte [0-3]
 * "x" = exit
 *
 ******************************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 * THE AUTHORS OF THIS PROGRAM ACCEPT ABSOLUTELY NO LIABILITY FOR
 * ANY HARM OR LOSS RESULTING FROM ITS USE.  IT IS _EXTREMELY_ UNWISE
 * TO RELY ON SOFTWARE ALONE FOR SAFETY.  Any machinery capable of
 * harming persons must have provisions for completely removing power
 * from all motors, etc, before persons enter any danger area.  All
 * machinery must be designed to comply with local and national safety
 * codes, and the authors of this software can not, and do not, take
 * any responsibility for such compliance.
 *
 * This code was written as part of the LinuxCNC project.  For more
 * information, go to www.linuxcnc.org.
 *
 ******************************************************************************/



#define  PC          // this is the PC-version
//#define  THREADED  // use threads
#define  WITH_SHMEM  // we use shmem

//#define WITH_RT_SHMEM // we use RT-shmem





#include <sys/time.h>
#include <termios.h>
#include <assert.h>
#include <pthread.h>
#include <sys/mman.h>
#include <limits.h>
#include <stdbool.h>
#include <signal.h>
#include <arpa/inet.h>
#include <linux/if_packet.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <fcntl.h>
#include <sys/io.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <stdio.h>
#include <string.h>
#include <signal.h>

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/ether.h>
#include <sys/ioctl.h>
#include <linux/if.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>
#include <arpa/inet.h>
#include <linux/udp.h>
#include <linux/ip.h>


void help(){
	printf("compile with (when using usersapce):\n");
	printf("\t gcc  pcIOethRaw.c -o pcIOethRaw -pthread\n");
	printf("\n");
	printf("\n");

	printf("compile with (when using RTAI-version):\n");
	printf("\t (#define WITH_RT_SHMEM // we use RT-shmem\n");
	printf("\t export LPATH=[path to linCNC-source]/\n");
	printf("\t gcc -Os -g -I. -I${LPATH}linuxcnc.2.7.14/include -I${LPATH}linuxcnc.2.7.14/src/include -DUSPACE -fno-fast-math -mieee-fp -fno-unsafe-math-optimizations -DRTAPI -D_GNU_SOURCE -Drealtime -D__MODULE__ -I${LPATH}linuxcnc.2.7.14/include -Wframe-larger-than=2560 -DSIM -fPIC -URTAPI -U__MODULE__ -DULAPI -Os  -o pcIOethRaw pcIOethRaw.c  -Wl,-rpath,${LPATH}linuxcnc.2.7.14/lib -L${LPATH}linuxcnc.2.7.14/lib -llinuxcnchal -lpthread\n");
	
	printf("\n");
	printf("\n");
}


#include "defs.h"
#include "commandline.c"

#include "func.c"


#ifdef WITH_RT_SHMEM
#include "rtapi.h"              /* user-level API to RT Linux */

int get_rtshmem()
{
    int retval;
    int shID;

    module = rtapi_init("ethsh");
    if (module < 1) {
        rtapi_print_msg(RTAPI_MSG_ERR,
            "shmemusr main: rtapi_init returned %d\n", module);
        return -1;
    }
    rtapi_set_msg_level(RTAPI_MSG_INFO);

    printf("ethsh: module: %i    shkey:%i\n", module,shkey);

    // allocate the shared memory structure 
    shID = rtapi_shmem_new(shkey, module, sizeof(rasport_t));
    if (shID < 0) {
        rtapi_print_msg(RTAPI_MSG_ERR,
            "shmemusr main: rtapi_shmem_new returned %d\n", shID);
        rtapi_exit(module);
        return -1;
    }
    printf("ethsh: hID: %i\n", shID);
    retval = rtapi_shmem_getptr(shID, (void **) &recbufs);
    if (retval < 0) {
        rtapi_print_msg(RTAPI_MSG_ERR,
            "shmemusr main: rtapi_shmem_getptr returned %d\n", retval);
        rtapi_exit(module);
        return -1;
    }
    printf("ethsh: retval: %i\n", retval);

	return retval;
}
#endif


static unsigned int  lastEthInNr=0, pktInNr=0;
static bool firstEthIn=false;


int rec_data(){
// receiving part
	bool doit=false;

        int count=0;

	if (!rec_raw()) return 0;  // this is not ours
//	return 0;
	{ // the packt ist now in recbufs->validoutval
	if (debug & deb_match)	print_header(numbytes);

//	if (debug>0)   print_buf(recbuf);
//	printf("recbuf[17]: %2x, ->outval[0]:%2x, inval[0]:%2x, recbuf[17+5]:%2x \n", recbuf[17],recbufs->outval[0],recbufs->inval[0],recbuf[17+5]);
	doit=lastin[0]!=recbufs->validoutval[0];
  	 if (lastin[1]!=recbufs->validoutval[1]) doit=true;
	 if (lastin[2]!=recbufs->validoutval[2]) doit=true;
	 if (lastin[3]!=recbufs->validoutval[3]) doit=true;
	 if (lastin[4]!=recbufs->validoutval[4]) doit=true;
         if (doit){
        	lastin[0]=recbufs->validoutval[0];
	        lastin[1]=recbufs->validoutval[0];
        	lastin[2]=recbufs->validoutval[2];
	        lastin[3]=recbufs->validoutval[3];
	        lastin[4]=recbufs->validoutval[4];
		pktInNr=recbufs->sndrawcount;       // this number calculated by  the _SENDER_
        if (firstEthIn) {firstEthIn=false;lastEthInNr=pktInNr-1;}
        if (debug&2048)
		 printf("lastEthInNr:%4x pktInNr:%4x pktInNr-1:%4x-1\n",lastEthInNr,pktInNr,((pktInNr-1)/*&255*/));


        if (lastEthInNr!=((pktInNr-1)/*&255*/)){printf("PACKET LOST!!!lastEthInNr:%4x pktInNr:%4x\n",lastEthInNr,pktInNr);}
        lastEthInNr=pktInNr;
        }
        count++;
        if (count>100000){
        print_stamp();


                printf(" 100k packtes received.\n");
                count=0;
                }
	}  // isForMe   
//        print_stamp();
	// end of receiving part
}


















void  *thread_func(void *data){
//////////////////////////////////////////////////////////////
	int lcount=0;
while (1==1) {
	usleep(1000);
	lcount++;
	if ((testmode & 1)==0)	if(snd_raw2raspi()) snd_udp();
	if ((testmode & 2)==0)  rec_data();
	if (testmode & 4){
	        port_data_array->inval[0]= port_data_array->inval[0]= ~port_data_array->inval[0]; // ^= 2;
	        port_data_array->outval[0]= port_data_array->outval[0]= ~port_data_array->outval[0]; // ^= 2;
		if (lcount>100000){lcount=0;print_stamp();printf(" 100k packets sent\n"); }	
	}
	}

//////////////////////////////////////////////////////////////////////////////////

}

void srtask() { // send and recieve task
//        if ((testmode & 1)==0)    if(snd_raw2raspi()) snd_udp();; // snd_raw2raspi();
        if ((testmode & 1)==0)   {snd_raw2raspi(); snd_udp();}; // snd_raw2raspi();
        if ((testmode & 2)==0)    rec_data();
        if (testmode & 4){
                        if (testmode & 8 ) printf("testmode 8\n");
//                         bcm2835_gpio_write(PINaS, HIGH);
//                       bcm2835_gpio_write(PINaS, LOW);
                }
	
//	rec_udp();   // this version receives only RAW-eth
}



int main(int argc, char *argv[])
{
printf("pcIOethRaw driver and testtool - running on pc-linuxcnc-side\n");
printf("29.3.2019 (c) by Ralf Burger\n");

printf("\n");
printf("\n");
	getCLopt(argc, argv);
printf("destination MAC: "MAC_STR"\n",
        smac[0],
        smac[1],
        smac[2],
        smac[3],
        smac[4],
        smac[5]);

        s_dstMac[0] =smac[0];
        s_dstMac[1] =smac[1];
        s_dstMac[2] =smac[2];
        s_dstMac[3] =smac[3];
        s_dstMac[4] =smac[4];
        s_dstMac[5] =smac[5];

printf("rec. for MAC: "MAC_STR"\n",
        rmac[0],
        rmac[1],
        rmac[2],
        rmac[3],
        rmac[4],
        rmac[5]);

        r_srcMac[0] =rmac[0];
        r_srcMac[1] =rmac[1];
        r_srcMac[2] =rmac[2];
        r_srcMac[3] =rmac[3];
        r_srcMac[4] =rmac[4];
        r_srcMac[5] =rmac[5];




	/* Get the index of the interface to send on */
	memset(&if_idx, 0, sizeof(struct ifreq));
	strncpy(if_idx.ifr_name, dev_name, IFNAMSIZ-1);
	printf("interface:%s\n",dev_name);


#ifdef WITH_RT_SHMEM   
	get_rtshmem();          // use rtai-shmem
#else
	get_shmem();           // use standard-shmem
#endif

memset(sndbuf, 0, BUF_SIZ);

snd_header->h_source[0] = ((uint8_t *)&if_mac.ifr_hwaddr.sa_data)[0];
snd_header->h_source[1] = ((uint8_t *)&if_mac.ifr_hwaddr.sa_data)[1];
snd_header->h_source[2] = ((uint8_t *)&if_mac.ifr_hwaddr.sa_data)[2];
snd_header->h_source[3] = ((uint8_t *)&if_mac.ifr_hwaddr.sa_data)[3];
snd_header->h_source[4] = ((uint8_t *)&if_mac.ifr_hwaddr.sa_data)[4];
snd_header->h_source[5] = ((uint8_t *)&if_mac.ifr_hwaddr.sa_data)[5];
snd_header->h_dest[0] = smac[0];
snd_header->h_dest[1] = smac[1];
snd_header->h_dest[2] = smac[2];
snd_header->h_dest[3] = smac[3];
snd_header->h_dest[4] = smac[4];
snd_header->h_dest[5] = smac[5];
snd_header->h_proto =  htons(ETH_P_IP);
tx_len += sizeof(struct ether_header)+2;

if (init_raw_rec_socket(dev_name)<0)exit(-1);
if (init_raw_snd_socket(dev_name)<0)exit(-1);
if (init_udp_snd()<0)exit(-1);
printf("udpssockfd: %i\n",udpssockfd);

//init_thread();  //no threads any more
char c='y';int mode=1;int byte=0;int f=0;
port_data_array->inval[0]=0;
printf("mode = OUT\n");
while (c!='x'){
	srtask();
	c = getch();
        if (c == 'i') {mode=0;printf("mode = IN\n");}
        if (c == 'o') {mode=1;printf("mode = OUT\n");}
        if (c == '+') {byte++; if (byte==4) byte=0; printf("IO-byte:%01d\n",byte);}



        for (f=1; f<=8; f++){
           if(c == 48+f){
                printf("(1<<f):=%2d\n",(1<<f-1));
                if (mode==0){
                        printf(" toggle IN bit %01d of byte[%01d]\n",f,byte);
                        printf("%2x:",port_data_array->inval[byte]);
                        port_data_array->inval[byte] ^= (1<<(f-1));
                        printf("->%2x\n",port_data_array->inval[byte]);
                        }
                if (mode==1){
                        printf(" toggle OUT bit %01d of byte[%01d]\n",f,byte);
                        printf("%2x:",port_data_array->outval[byte]);
                        port_data_array->outval[byte] ^= (1<<(f-1));
                        printf("->%2x\n",port_data_array->outval[byte]);
                        }
                }
             }
}


	return 0;
}





